/* memory allocation routines
 *
 * Adapted from alloc routine in K&R; memory statistics and interrupt
 * protection added for use with net package. Must be used in place of
 * standard Turbo-C library routines because the latter check for stack/heap
 * collisions. This causes erroneous failures because process stacks are
 * allocated off the heap.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/errno.h>
int free(); 		/* This must be before we include global.h */
int (*xxfree)() = free; /* if free() does not check for NULL pointers */
#include "global.h"
#include "config.h"
#include "memory.h"
#include "sems.h"

/**** SMALL DATA MODEL VERSION ****/

HEADER *morecore();
void shfree();
unsigned coreleft();

struct memory *Mem; /* Pointer to the current memory segment */

/* Allocate block of 'nb' bytes */
char *
shmalloc(nb)
unsigned nb;
{
	register HEADER *p, *q;
	register unsigned nu;
	struct sembuf sembuf;
	struct mrec *mrec;
	char i_state;

	if (Mem == NULLMEM)
	    return malloc(nb);
	if(nb == 0)
		return NULL;
	i_state = dirps();
	mrec = (struct mrec *) Mem->shmbase;
	/* Round up to full block, then add one for header */
	nu = ((nb + ABLKSIZE - 1) / ABLKSIZE) + 1;
loop:
	if ((q = *Mem->allocp) == NULLHDR){
		Mem->base->s.ptr = *Mem->allocp = q = Mem->base;
		Mem->base->s.size = 1;
	}
	for (p = q->s.ptr; ; q = p, p = p->s.ptr){
		if (p->s.size >= nu){
			/* This chunk is at least as large as we need */
			if (p->s.size <= nu + 1){
				/* This is either a perfect fit (size == nu)
				 * or the free chunk is just one unit larger.
				 * In either case, alloc the whole thing,
				 * because there's no point in keeping a free
				 * block only large enough to hold the header.
				 */
				q->s.ptr = p->s.ptr;
			} else {
				/* Carve out piece from end of entry */
				p->s.size -= nu;
				p += p->s.size;
				p->s.size = nu;
			}
			*Mem->allocp = q;
			p->s.ptr = p;	/* for auditing */
			mrec->allocs++;
			restore(i_state);
			return (char *)(p + 1);
		}
		if (p == *Mem->allocp && (p = morecore(nu)) == NULLHDR){
			mrec->memfail++;
			if (Mem->block) { /* Block until there is room */
				sembuf.sem_num = SEM_FREE;
				sembuf.sem_flg = IPC_NOWAIT;
				sembuf.sem_op = WAITDEC_SEM;
				semop(Mem->semid, &sembuf, 1); /* set to 0 */
				sembuf.sem_flg = SEM_UNDO;
				restore(i_state);
				while(semop(Mem->semid, &sembuf, 1) == -1
				      && errno == EINTR) ;
				goto loop; /* quick hack */
			}
			restore(i_state);
			return NULL;
		}
	}
}
/* Get more memory from the system and put it on the heap */
HEADER *
morecore(nu)
unsigned nu;
{
	register char *cp;
	register HEADER *up;

	return NULLHDR;
}
/* Grab the specified amount of memory from the system,
 * and place it on the heap
 */
unsigned long
grabcore(m, size, addr, key)
struct memory *m;
unsigned long size; /* Only the server should specify a size other than 0 */
int32 addr;
key_t key;
{
	register HEADER *up;
	char *shmat();
	struct shmid_ds buf;
	struct mrec *mrec;
	/* When size equals 0 a previously created segment is attached */
	if ((m->shmid = shmget(key, size, size ? (0666 | IPC_CREAT) :0)) == -1)
	    perror("shmget");

	m->shmbase = shmat(m->shmid, addr, 0);

	shmctl(m->shmid, IPC_STAT, &buf);
	m->heapsize = buf.shm_segsz - sizeof(struct mrec);
	m->heapbase = m->shmbase + sizeof(struct mrec);
	m->heaptop = m->heapbase + m->heapsize;
	mrec = (struct mrec *)m->shmbase;
	m->base = (HEADER *) &mrec->base;
	m->allocp = (HEADER **) &mrec->allocp;

	/* Initialize the heap pointers */
	if (size){
		m->base->s.ptr = *m->allocp = m->base;
		m->base->s.size = 1;
		up = (HEADER *)m->heapbase;
		up->s.size = m->heapsize / sizeof(HEADER);
		up->s.ptr = up;	/* satisfy audit */
		activatecore(m);
		shfree((void *)(up + 1));
	}
	else activatecore(m);
	return m->heapsize;
}

activatecore(m)
struct memory *m;
{
    Mem = m;
}

/* Put memory block back on heap */
void
shfree(blk)
void *blk;
{
	register HEADER *p,*q;
	unsigned short *ptr;
	struct mrec *mrec;
	struct sembuf sembuf;
	char i_state;

	if(blk == NULL)
		return;		/* Required by ANSI */
	if(Mem == NULLMEM) {
		free(blk);
		return;
	}
	i_state = dirps();
	mrec = (struct mrec *) Mem->shmbase;
	p = (HEADER *)blk - 1;
	/* Audit check */
	if(p->s.ptr != p){
		ptr = (unsigned short *)&blk;
		printf("WARNING!! freeing garbage (0x%lx) pc = 0x%x %x\n",ptol(blk),
			ptr[-1],ptr[-2]);
		mrec->garbage++;
		restore(i_state);
		return;
	}
	/* Search the free list looking for the right place to insert */
	for(q = *Mem->allocp; !(p > q && p < q->s.ptr); q = q->s.ptr){
		/* Highest address on circular list? */
		if(q >= q->s.ptr && (p > q || p < q->s.ptr))
			break;
	}
	if(p + p->s.size == q->s.ptr){
		/* Combine with front of this entry */
		p->s.size += q->s.ptr->s.size;
		p->s.ptr = q->s.ptr->s.ptr;
	} else {
		/* Link to front of this entry */
		p->s.ptr = q->s.ptr;
	}
	if(q + q->s.size == p){
		/* Combine with end of this entry */
		q->s.size += p->s.size;
		q->s.ptr = p->s.ptr;
	} else {
		/* Link to end of this entry */
		q->s.ptr = p;
	}
	*Mem->allocp = q;
	mrec->frees++;
	restore(i_state);
	/* Signal anybody that might be waiting */
	sembuf.sem_num = SEM_FREE;
	sembuf.sem_flg = IPC_NOWAIT;
	sembuf.sem_op = TEST0_SEM;
	if(semop(Mem->semid, &sembuf, 1) == 0) {
	    sembuf.sem_op = INC_SEM;
	    semop(Mem->semid, &sembuf, 1);
	}
}

#ifdef	notdef	/* Not presently used */
/* Move existing block to new area */
void *
shrealloc(area,size)
void *area;
unsigned size;
{
	unsigned osize;
	HEADER *hp;
	char *cp;
	char i_state;

	hp = ((HEADER *)area) - 1;
	osize = (hp->s.size -1) * ABLKSIZE;

	/* Make sure nobody else comes in and takes it */
	i_state = dirps();
	free(area);
	if((cp = malloc(size)) != NULL && cp != area)
		memcpy((char *)cp,(char *)area,size>osize? osize : size);
	restore(i_state);
	return cp;
}
#endif
/* Allocate block of cleared memory */
void *
shcalloc(nelem,size)
unsigned nelem;	/* Number of elements */
unsigned size;	/* Size of each element */
{
	register unsigned i;
	register char *cp;

	i = nelem * size;
	if((cp = shmalloc(i)) != NULL)
		memset(cp,0,i);
	return (void *)cp;
}
/* Print free list map */
memstat()
{
	HEADER *p;
	struct mrec *mrec;    
	unsigned long total = 0;
	int i = 0;
	char i_state;

	i_state = dirps();
	mrec = (struct mrec *) Mem->shmbase;
	for(p = Mem->base->s.ptr;p != Mem->base;p = p->s.ptr){
		total += p->s.size * sizeof(HEADER);
		printf("%5lx %6lu",ptol(p),(long)p->s.size * ABLKSIZE);
		if(++i == 4)
			putchar('\n');
		else
			printf(" | ");
	}
	if(i != 0)
		putchar('\n');
	printf("heap base %lx size %lu avail %lu (%lu%%) coreleft %u\n",
	       ptol((void *)Mem->heapbase),Mem->heapsize,total,
	       100L*total/Mem->heapsize,coreleft());
	printf("allocs %lu frees %lu (diff %lu) mfails %lu garbage %lu\n",
	       mrec->allocs,mrec->frees,mrec->allocs - mrec->frees,
	       mrec->memfail,mrec->garbage);
	restore(i_state);
}
