/* TCP implementation. Follows RFC 793 as closely as possible */
#ifndef	NULLTCB

#include "global.h"
#include "netuser.h"
#include "timer.h"

#define	DEF_WND	2048	/* Default receiver window */
#define	NTCB	19	/* # TCB hash table headers */
#define	RTTCACHE 16	/* # of TCP round-trip-time cache entries */
#define	DEF_MSS	512	/* Default maximum segment size */
#define	DEF_RTT	5000	/* Initial guess at round trip time (5 sec) */
#define	MSL2	30	/* Guess at two maximum-segment lifetimes */

extern int32 Clock;
#define	geniss()	((int32)Clock << 18)	/* Increment clock at 4.5 Mbytes/sec */

/* Round trip timing parameters */
#define	AGAIN	8	/* Average RTT gain = 1/8 */
#define	LAGAIN	3	/* Log2(AGAIN) */
#define	DGAIN	4	/* Mean deviation gain = 1/4 */
#define	LDGAIN	2	/* log2(DGAIN) */

/* TCP segment header -- internal representation
 * Note that this structure is NOT the actual header as it appears on the
 * network (in particular, the offset and checksum fields are missing).
 * All that knowledge is in the functions ntohtcp() and htontcp() in tcpsubr.c
 */
struct tcp {
	int16 source;	/* Source port */
	int16 dest;	/* Destination port */
	int32 seq;	/* Sequence number */
	int32 ack;	/* Acknowledgment number */
	struct {
		char urg;
		char ack;
		char psh;
		char rst;
		char syn;
		char fin;
	} flags;
	int16 wnd;	/* Receiver flow control window */
	int16 up;	/* Urgent pointer */
	int16 mss;	/* Optional max seg size */
};
/* TCP options */
#define	EOL_KIND	0
#define	NOOP_KIND	1
#define	MSS_KIND	2

#define	TCPLEN		20
#define	MSS_LENGTH	4
/* Resequencing queue entry */
struct reseq {
	struct reseq *next;	/* Linked-list pointer */
	struct tcp seg;		/* TCP header */
	struct mbuf *bp;	/* data */
	int16 length;		/* data length */
	char tos;		/* Type of service */
};
#define	NULLRESEQ	(struct reseq *)0

/* TCP connection control block */
struct tcb {
	struct tcb *prev;	/* Linked list pointers for hash table */
	struct tcb *next;

	struct connection conn;

	char state;	/* Connection state */
#define	CLOSED		0	/* Must be 0 */
#define	LISTEN		1
#define	SYN_SENT	2
#define	SYN_RECEIVED	3
#define	ESTABLISHED	4
#define	FINWAIT1	5
#define	FINWAIT2	6
#define	CLOSE_WAIT	7
#define	CLOSING		8
#define	LAST_ACK	9
#define	TIME_WAIT	10

	char reason;		/* Reason for closing */
#define	NORMAL		0	/* Normal close */
#define	RESET		1	/* Reset by other end */
#define	TIMEOUT		2	/* Excessive retransmissions */
#define	NETWORK		3	/* Network problem (ICMP message) */

/* If reason == NETWORK, the ICMP type and code values are stored here */
	char type;
	char code;

	/* Send sequence variables */
	struct {
		int32 una;	/* First unacknowledged sequence number */
		int32 nxt;	/* Next sequence num to be sent for the first time */
		int32 ptr;	/* Working transmission pointer */
		int32 wl1;	/* Sequence number used for last window update */
		int32 wl2;	/* Ack number used for last window update */
		int16 wnd;	/* Other end's offered receive window */
		int16 up;	/* Send urgent pointer */
	} snd;
	int32 iss;		/* Initial send sequence number */
	int32 resent;		/* Count of bytes retransmitted */
	int16 cwind;		/* Congestion window */
	int16 ssthresh;		/* Slow-start threshold */

	/* Receive sequence variables */
	struct {
		int32 nxt;	/* Incoming sequence number expected next */
		int16 wnd;	/* Our offered receive window */
		int16 up;	/* Receive urgent pointer */
	} rcv;
	int32 irs;		/* Initial receive sequence number */
	int32 rerecv;		/* Count of duplicate bytes received */
	int16 mss;		/* Maximum segment size */

	int16 window;		/* Receiver window and send queue limit */

	void (*r_upcall)();	/* Call when "significant" amount of data arrives */
	void (*t_upcall)();	/* Call when ok to send more data */
	void (*s_upcall)();	/* Call when connection state changes */
	struct {		/* Control flags */
		char force;	/* We owe the other end an ACK or window update */
		char clone;	/* Server-type TCB, cloned on incoming SYN */
		char retran;	/* A retransmission has occurred */
		char active;	/* TCB created with an active open */
		char synack;	/* Our SYN has been acked */
		char rtt_run;	/* We're timing a segment */
	} flags;
	char tos;		/* Type of service (for IP) */
	char backoff;		/* Backoff interval */

	struct mbuf *rcvq;	/* Receive queue */
	struct mbuf *sndq;	/* Send queue */
	int16 rcvcnt;		/* Count of items on rcvq */
	int16 sndcnt;		/* Number of unacknowledged sequence numbers on
				 * sndq. NB: includes SYN and FIN, which don't
				 * actually appear on sndq!
				 */

	struct reseq *reseq;	/* Out-of-order segment queue */
	struct timer timer;	/* Retransmission timer */
	int32 rtt_time;		/* Stored clock values for RTT */
	int32 rttseq;		/* Sequence number being timed */
	int32 srtt;		/* Smoothed round trip time, milliseconds */
	int32 mdev;		/* Mean deviation, milliseconds */

	int user;		/* User parameter (e.g., for mapping to an
				 * application control block
				 */
};
#define	NULLTCB	(struct tcb *)0
/* TCP round-trip time cache */
struct tcp_rtt {
	int32 addr;		/* Destination IP address */
	int32 srtt;		/* Most recent SRTT */
	int32 mdev;		/* Most recent mean deviation */
};
#define	NULLRTT	(struct tcp_rtt *)0
extern struct tcp_rtt Tcp_rtt[];

/* TCP statistics counters */
struct tcp_stat {
	int16 runt;		/* Smaller than minimum size */
	int16 checksum;		/* TCP header checksum errors */
	int16 conout;		/* Outgoing connection attempts */
	int16 conin;		/* Incoming connection attempts */
	int16 resets;		/* Resets generated */
	int16 bdcsts;		/* Bogus broadcast packets */
};
extern struct tcp_stat Tcp_stat;
extern struct tcb *Tcbs[];
extern int16 Tcp_mss;
extern int16 Tcp_window;
extern int32 Tcp_irtt;

#if	defined(__STDC__) || defined(__TURBOC__)
void tcp_input(struct mbuf *bp,struct ip *ip,int rxbroadcast);
void tcp_icmp(int32 source,int32 dest,char type,char code,struct mbuf **bpp);
struct tcb *lookup_tcb(struct connection *conn);
struct tcb *create_tcb(struct connection *conn);
void tcp_output(struct tcb *tcb);
void close_self(struct tcb *tcb,int reason);
void setstate(struct tcb *tcb,int newstate);
struct mbuf *htontcp(struct tcp *tcph,struct mbuf *data,
	struct pseudo_header *ph);
int ntohtcp(struct tcp *tcph,struct mbuf **bpp);
int seq_within(int32 x,int32 low,int32 high);
int seq_lt(int32 x,int32 y);
int seq_le(int32 x,int32 y);
int seq_gt(int32 x,int32 y);
int seq_ge(int32 x,int32 y);
void link_tcb(struct tcb *tcb);
void unlink_tcb(struct tcb *tcb);
void send_syn(struct tcb *tcb);
/* TCP primitives */
struct tcb *open_tcp(struct socket *lsocket,struct socket *fsocket,
	int mode,int16 window,void (*r_upcall)(),void (*t_upcall)(),
	void (*s_upcall)(),int tos,int user);
int send_tcp(struct tcb *tcb,struct mbuf *bp);
int recv_tcp(struct tcb *tcb,struct mbuf **bpp,int16 cnt);
int close_tcp(struct tcb *tcb);
int del_tcp(struct tcb *tcb);
int tcpval(struct tcb *tcb);
int kick_tcp(struct tcb *tcb);
void st_tcp(struct tcb *tcb);
void tcp_dump(struct mbuf **bpp,int32 source,int32 dest,int check);
void reset_tcp(struct tcb *tcb);
char *tcp_port(int16 n);
void rtt_add(int32 addr,int32 rtt);
struct tcp_rtt *rtt_get(int32 addr);
void reset(struct ip *ip,struct tcp *seg);
void reset_all(void);

#else

struct tcb *lookup_tcb();
struct tcb *create_tcb();
void tcp_output(),tcp_input(),close_self(),setstate();
struct mbuf *htontcp();
int ntohtcp();
int seq_within(),seq_lt(),seq_le(),seq_gt(),seq_ge();
void link_tcb(),unlink_tcb();
void tcp_icmp();
void send_syn();
/* TCP primitives */
struct tcb *open_tcp();
int send_tcp(),recv_tcp(),close_tcp(),del_tcp();
int tcpval();
void st_tcp(),tcp_dump();
int kick_tcp();
void reset_tcp();
char *tcp_port();
void rtt_add();
struct tcp_rtt *rtt_get();
void reset();
void reset_all();
#endif

#endif	/* NULLTCB */
